import java.io.*;

/** Komprimiert Daten mittels des LZ77 Algorithmusses */
public class LZ77 {
	private SlidingByteWindow	lookahead, data;	// Die Gleitfenster
	private ByteElementPool		pool;				// Der Elementpool
	
	/**
	* Konstruktor.
	* @param datasize Gre des Datenfensters.
	* @param lookaheadsize Gre des Lookahead-Fensters.
	*/
	public LZ77( int datasize, int lookaheadsize ){
		this.pool		= new ByteElementPool();
		this.lookahead	= new SlidingByteWindow( lookaheadsize, this.pool );
		this.data		= new SlidingByteWindow( datasize, this.pool );
	}

	/**
	* Komprimiert einen Datenstrom.
	* @param in Zu komprimierender Datenstrom.
	* @param out Komprimierter Ausgabedatenstrom.
	* @throws IOException wenn in den Datenstrmen ein IO-Fehler auftritt.
	*/
	public void encode( InputStream in, OutputStream out ) throws IOException {
		int					readamount, shiftlength;
		byte[]				readbuffer;
		int[]				index_and_length;
		DataOutputStream	dout;
		ByteElement			overflow;
		
		if( in == null || out == null )
			throw new IOException( "input and/or output stream is null" );
		
		dout				= new DataOutputStream( out );
		readbuffer			= new byte[lookahead.getSize()];
		index_and_length	= new int[2];
			
		readamount = in.read( readbuffer, 0, 1 );
		if( readamount != 1 ) throw new IOException( "no data available" );
		
		data.shiftIn( readbuffer, 0, 1 );
		
		// erstes byte encoden
		dout.writeByte( (int)readbuffer[0] );
		dout.writeShort( 0 );
		dout.writeShort( 0 );
		
		readamount = in.read( readbuffer );
		if( readamount > 0 ) lookahead.shiftIn( readbuffer, 0, readamount );
		
		while( lookahead.getFillsize() > 0 ){
			// index und laenge der groessten uebereinstimmung ermitteln
			data.findGreatestCommonOccurance( lookahead, index_and_length );
			
			// die laenge inklusive des zusaetzlichen bytes ins datenfenster shiften
			shiftlength = index_and_length[1] + 1;
			overflow = lookahead.shiftOut( shiftlength );
			overflow = data.shiftIn( overflow );
			this.pool.add( overflow );
			
			//neue bytes ins lookaheadfenster lesen
			readamount = in.read( readbuffer, 0, lookahead.getSize() - lookahead.getFillsize() );
			if( readamount > 0 ){
				overflow = lookahead.shiftIn( readbuffer, 0, readamount );
				overflow = data.shiftIn( overflow );
				this.pool.add( overflow );
			}
			
			// Tupel encoden
			dout.writeByte( (int)data.getLast() );
			dout.writeShort( index_and_length[0] );
			dout.writeShort( index_and_length[1] );
		}
	}
	
	/**
	* Dekomprimiert einen Datenstrom.
	* ACHTUNG: Es wird KEINE berprfung der Eingabedaten durchgefhrt!
	* Wenn es sich bei den Eingabedaten NICHT um LZ77 komprimierte Daten handelt
	* merkt man das frhestens daran, dass die Methode bei Problemen irgendeine
	* Form der RuntimeException wirft.
	* @param in Eingabedatenstrom.
	* @param out Ausgabedatenstrom mit dekomprimierten Daten.
	* @throws IOException wenn ein IO-Fehler in den Datenstrmen auftritt.
	*/
	public void decode( InputStream in, OutputStream out ) throws IOException {
		int				readamount, index, length;
		byte[]			readbuffer;
		DataInputStream	din;
		ByteElement		transfer, pos;
		
		if( in == null || out == null )	
			throw new IOException( "input and/or output Stream is null" );
		
		din = new DataInputStream( in );
		readbuffer = new byte[1];
		
		// das erste codierte zeichen einlesen. index und laenge ignorieren, sind 0.
		readamount = din.read( readbuffer, 0, 1 );
		if( readamount != 1 ) throw new IOException( "no data available" );
		din.readShort();
		din.readShort();
		
		data.shiftIn( readbuffer, 0, 1 );
		
		// zeichen ausgeben
		out.write( readbuffer, 0, 1 );
		
		while( din.read(readbuffer, 0, 1) == 1 ){
			index = (int)din.readShort();
			length = (int)din.readShort();
			
			// decodierten kram bestimmen, ...
			transfer = data.copyElements( index, length );
			
			// ...ausgeben...
			pos = transfer;
			while( pos != null ){
				out.write( (int)pos.entry );
				pos = pos.next;
			}
			out.write( (int)readbuffer[0] );

			// ...und ins datenfenster shiften
			transfer = data.shiftIn( transfer );
			this.pool.add( transfer );
			transfer = data.shiftIn( readbuffer, 0, 1 );
			this.pool.add( transfer );
		}
 	}
}