/** Implementiert ein Gleitfenster fr den LZ77 Algorithmus */
public class SlidingByteWindow {
	public static final int MAX_WINDOWSIZE = 32768;	// Maximale Fenstergre
	
	private ByteElement		first, last;	// erster und letzter Fenstereintrag
	private int				size, fillsize;	// Max. und derzeitige Gre des Fensters
	private ByteElementPool	pool;			// Der verwendete Elementpool.
	
	/**
	* Konstruktor.
	* @param size Die Gre dieses Gleitfensters.
	* @param pool Der zu verwendende Elementpool.
	* @throws IndexOutOfBoundsException wenn size > MAX_WINDOWSIZE oder size < 0.
	* @throws NullPointerException wenn pool null ist.
	*/
	public SlidingByteWindow( int size, ByteElementPool pool ){
		if( size > MAX_WINDOWSIZE )
			throw new IndexOutOfBoundsException( "max windowsize is "+MAX_WINDOWSIZE );
		
		if( size < 0 )
			throw new IndexOutOfBoundsException( "size has to be > 0" );
		
		if( pool == null )
			throw new NullPointerException( "pool may not be null" );
			
		this.first		= null;
		this.last		= null;
		this.fillsize	= 0;
		this.size		= size;
		this.pool		= pool;
	}
	
	/**
	* Gibt die maximale Gre dieses Gleitfensters zurck.
	* @returns die maximale Gre dieses Gleitfensters.
	*/
	public int getSize(){ return this.size; }
	
	/**
	* Gibt die derzeitige Gre dieses Gleitfensters zurck.
	* @returns die derzeitige Gre dieses Gleitfensters.
	*/
	public int getFillsize(){ return this.fillsize; }
	
	/**
	* Gibt einen Eintrag in diesem  Gleitfenster zurck.
	* @param index Die Position des eintrags im Fenster.
	* @returns Der byte-Wert des angegebenen Eintrags.
	* @trhows IndexOutOfBoundsException wenn der gewnschte Index ausserhalb der Fensterdaten liegt.
	*/
	public byte get( int index ){
		if( index < 0 || index > fillsize )
			throw new IndexOutOfBoundsException();
		
		ByteElement pos = this.first;
		for( int i = 0; i < index; i++ )
			pos = pos.next;
			
		return pos.entry;
	}
	
	/**
	* Gibt den letzten Eintrag dieses Gleitfensters zurck.
	* @returns Der byte-Wert des letzten Eintrags des Fensters.
	*/
	public byte getLast(){ return last.entry; }
	
	/**
	* Gibt das erste Element dieses Gleitfensters
	* (bzw. die vollstndige Liste seiner Elemente) zurck.
	* @returns Das Erste Element der Elmentliste des Fensters.
	*/
	protected ByteElement getFirstElement(){ return this.first; }
	
	/**
	* Gleitfensteroperation, die von "rechts" Daten aus einem Array
	* in das Fenster schiebt. Ggf. werden dabei links Daten herausgeschoben.
	* @param b Das Array mit den Daten, die in das Fenster geschoben werden sollen.
	* @param off Das Startoffset im Array b.
	* @param length Die Lnge der Daten aus dem Array b, die ins Fenster geschoben werden sollen.
	* @returns Eine Liste mit Elementen, die bei der Operation "links" aus dem Fenster herausgeschoben wurden.
	* @throws IndexOutOfBoundsException wenn off und/oder length nicht zur Lnge des Arrays b passen.
	*/
	public ByteElement shiftIn( byte[] b, int off, int length ){
		int end = off + length;
		
		if( off < 0 || length < 0 || end > b.length )
			throw new IndexOutOfBoundsException();
			
		if( length == 0 ) return null;
		
		int i = off;	
		if( this.first == null ){
			this.first = pool.get();
			this.first.entry = b[i];
			this.last = this.first;
			i++;
		}
	
		for( ; i < end; i++ ){
			last.next = pool.get();
			last.next.entry = b[i];
			last = last.next;
		}
		
		this.fillsize += length;
		
		return this.truncate();
	}
	
	/**
	* Gleitfensteroperation, die von "rechts" eine Liste mit Elementen
	* in das Fenster schiebt. Ggf. werden dabei links Daten herausgeschoben.
	* @param e Das erste Element der Elementliste, die in das Fenster geschoben werden soll.
	* @returns Eine Liste mit Elementen, die bei der Operation "links" aus dem Fenster herausgeschoben wurden.
	*/
	public ByteElement shiftIn( ByteElement e ){
		if( e == null ) return null;
		
		if( this.first == null ){
			this.first = e;
			this.last = e;
		}
		else this.last.next = e;
		
		while( this.last.next != null ){
			this.last = this.last.next;
			this.fillsize++;
		}
		
		return this.truncate();
	}
	
	/**
	* Hilfsoperation der ShiftIn-Operationen, die "links" Elemente von der
	* Elementliste dieses Glietfensters abschneidet, falls seine maximale Gre
	* berschritten wird.
	* @returns Das erste Element der Elementliste, die "abgeschnitten" wurde. null falls keine.
	*/
	private ByteElement truncate(){
		if( this.size >= this.fillsize ) return null;
		
		int overflow = this.fillsize - this.size - 1;
		ByteElement begin	= this.first;
		ByteElement end		= this.first;
		for( int i = 0; i < overflow; i++ ) end = end.next;
		this.first = end.next;
		end.next = null;
		this.fillsize = this.size;
		return begin;
	}
	
	/**
	* Schiebt eine Anzahl von Elementen "links" aus diesem Gleitfenster hinaus.
	* @param length Die Anzahl der elemente, die hinaus geschoben werden sollen.
	* @returns Das erste Element der hinausgeschobenen Elementliste. null wenn keine.
	* @throws IndexOutOfBoundsException wenn mehr Elemente herausgeschoben werden sollen, als im Fenster vorhanden sind.
	*/
	public ByteElement shiftOut( int length ){
		if( length > this.fillsize )
			throw new IndexOutOfBoundsException();
			
		if( length <= 0 ) return null;
			
		ByteElement begin	= this.first;
		ByteElement end		= this.first;
		for( int i = 0; i < length - 1; i++ )
			end = end.next;
		this.first	= end.next;
		end.next	= null;
		this.fillsize -= length;
		return begin;
	}
	
	/**
	* Grundlegende LZ77 Operation, die die lngste bereinstimmung von Daten am
	* ANFANG (!) eines anderen Gleitfensters mit Daten IRGENDWO (!) in diesem
	* Gleitfenster bestimmt.
	* @param with Das andere Gleitfenster.
	* @param index_and_length Call-By-Reference-Parameter, der nach Abschluss der Operation an Position 0 den Index und an Position 1 die Lnge der lngsten bereinstimmung enthlt.
	* @throws IndexOutOfBoundsException wenn index_and_length nicht mindestens 2 Eintrge hat.
	*/
	public void findGreatestCommonOccurance( SlidingByteWindow with, int[] index_and_length ){
		if( index_and_length.length < 2 )
			throw new IndexOutOfBoundsException( "index_and_length has to be of size 2 or higher" );
		
		ByteElement occurance = with.getFirstElement();
		ByteElement search, wpos, epos;
		int index, length, bestindex, bestlength;
		
		bestindex	= 0;
		bestlength	= 0;
		index		= 0;
		search		= this.first;
		
		while( true ){
			while( search != null && search.entry != occurance.entry ){
				index++;
				search = search.next;
			}
	
			if( search == null ) break;
	
			wpos = search.next;
			epos = occurance.next;
			length = 1;
			while( wpos != null && epos != null && wpos.entry == epos.entry ){
				length++;
				wpos = wpos.next;
				epos = epos.next;
			}
			
			// die erste uebereinstimmung der laenge with.getFillsize() - 1
			// ist das maximal codierbare und daher ausreichend.
			
			int max_codeable = with.getFillsize() - 1;
			if( length >= max_codeable ){
				index_and_length[0]	= index;
				index_and_length[1]	= max_codeable;
				return;
			}
	
			if( length > bestlength ){
				bestindex	= index;
				bestlength	= length;
			}
			
			index++;
			search = search.next;
		}
		
		index_and_length[0] = bestindex;
		index_and_length[1] = bestlength;
	}
	
	/**
	* Kopiert eine Anzahl von Elementen aus diesem Gleitfenster.
	* @param index Die Position im Fenster, ab der die Elemente kopiert werden sollen.
	* @param length Die Anzahl der Elemente, die kopiert werden sollen.
	* @returns Das Erste Element der Liste der kopierten Elemente. null wenn length 0 ist.
	* @throws IndexOutOfBoundsException wenn index und/oder length die derzeitige Gre des Fensters ber- oder unterschreiten.
	*/
	public ByteElement copyElements( int index, int length ){
		if( index < 0 || length < 0 || index + length > this.fillsize )
			throw new IndexOutOfBoundsException();
			
		if( length == 0 ) return null;
		
		ByteElement pos = this.first;
		int i;
		for( i = 0; i < index; i++ )
			pos = pos.next;
		
		ByteElement first	= this.pool.get();
		ByteElement last	= first;
		first.entry			= pos.entry;
		
		for( i = 1; i < length; i++ ){
			pos			= pos.next;
			last.next	= this.pool.get();
			last		= last.next;
			last.entry	= pos.entry;
		}
		
		return first;
	}
}